@@ -11,9 +11,10 @@
 
 raise "sorry, 1.8 was last decade" unless RUBY_VERSION >= '1.9'
 
-# this version also works with kramdown 0.12, 0.13, 0.14, but 1.x has the right license
-gem 'kramdown', '~> 1.4.0'
+gem 'kramdown', '~> 1.8.0'
 require 'kramdown'
+my_span_elements =  %w{list figure xref eref iref cref spanx vspace}
+Kramdown::Parser::Html::Constants::HTML_SPAN_ELEMENTS.concat my_span_elements
 
 require 'rexml/parsers/baseparser'
 
@@ -26,11 +27,13 @@ end
 module Kramdown
 
   module Parser
+
     class RFC2629Kramdown < Kramdown
 
       def initialize(*doc)
         super
         @span_parsers.unshift(:xref)
+        @span_parsers.unshift(:iref)
       end
 
       XREF_START = /\{\{(.*?)\}\}/u
@@ -47,6 +50,17 @@ module Kramdown
       end
       define_parser(:xref, XREF_START, '{{')
 
+      IREF_START = /\(\(\((.*?)\)\)\)/u
+
+      # Introduce new (((target))) syntax for irefs
+      def parse_iref
+        @src.pos += @src.matched_size
+        href = @src[1]
+        el = Element.new(:iref, nil, {'target' => href}) # XXX
+        @tree.children << el
+      end
+      define_parser(:iref, IREF_START, '\(\(\(')
+
     end
   end
 
@@ -88,6 +102,7 @@ module Kramdown
         super
         @sec_level = 1
         @in_dt = 0
+        @footnote_names_in_use = {}
       end
 
       def convert(el, indent = -INDENTATION, opts = {})
@@ -155,19 +170,35 @@ module Kramdown
           }.compact.join("\n")
           "#{' '*indent}<t>#{result}</t>\n"
         else
+          artwork_attr = {}
           if blockclass
-            $stderr.puts "*** Unimplemented block class: #{blockclass}"
+            classes = blockclass.split(' ')
+            classes.each do |cl|
+              if md = cl.match(/\Alanguage-(.*)/)
+                artwork_attr["type"] = md[1] # XXX overwrite
+              else
+                $stderr.puts "*** Unimplemented block class: #{cl}"
+              end
+            end
           end
           # compensate for XML2RFC idiosyncracy by insisting on a blank line
           unless el.attr.delete('tight')
             result[0,0] = "\n" unless result[0,1] == "\n"
           end
-          "#{' '*indent}<figure#{el_html_attributes(el)}><artwork><![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]></artwork></figure>\n"
+          el.attr.each do |k, v|
+            if md = k.match(/\Aartwork-(.*)/)
+              el.attr.delete(k)
+              artwork_attr[md[1]] = v
+            end
+          end
+          "#{' '*indent}<figure#{el_html_attributes(el)}><artwork#{html_attributes(artwork_attr)}><![CDATA[#{result}#{result =~ /\n\Z/ ? '' : "\n"}]]></artwork></figure>\n"
         end
       end
 
       def convert_blockquote(el, indent, opts)
-        "#{' '*indent}<t><list style='empty'#{el_html_attributes(el)}>\n#{inner(el, indent, opts)}#{' '*indent}</list></t>\n"
+        text = inner(el, indent, opts)
+        text = "<t></t>" unless text =~ /</ # empty block quote
+        "#{' '*indent}<t><list style='empty'#{el_html_attributes(el)}>\n#{text}#{' '*indent}</list></t>\n"
       end
 
       def end_sections(to_level, indent)
@@ -183,6 +214,18 @@ module Kramdown
         end
       end
 
+      def clean_pcdata(parts)    # hack, will become unnecessary with XML2RFCv3
+        clean = ''
+        irefs = ''
+        # warn "clean_parts: #{parts.inspect}"
+        parts.each do |p|
+          md = p.match(%r{([^<]*)(.*)})
+          clean << md[1]
+          irefs << md[2]        # breaks for spanx... don't emphasize in headings!
+        end
+        [clean, irefs]
+      end
+
       def convert_header(el, indent, opts)
         # todo: handle appendix tags
         el = el.deep_clone
@@ -190,8 +233,9 @@ module Kramdown
         if options[:auto_ids] && !el.attr['anchor']
           el.attr['anchor'] = saner_generate_id(el.options[:raw_text])
         end
-        el.attr['title'] = inner(el, indent, opts)
-        "#{end_sections(el.options[:level], indent)}#{' '*indent}<section#{@sec_level += 1; el_html_attributes(el)}>\n"
+        clean, irefs = clean_pcdata(inner_a(el, indent, opts))
+        el.attr['title'] = clean
+        "#{end_sections(el.options[:level], indent)}#{' '*indent}<section#{@sec_level += 1; el_html_attributes(el)}>#{irefs}\n"
       end
 
       def convert_hr(el, indent, opts) # misuse for page break
@@ -303,15 +347,21 @@ module Kramdown
         if alignment = opts[:table_alignment]
           alignment = alignment.shift
           if cols = opts[:table_cols].shift
-            md = cols.match(/(\d*)(.*)/)
-            widthopt = "width='#{md[1]}' " if md[1].to_i != 0
-            alignment = COLS_ALIGN[md[2]] || :left
+            md = cols.match(/(\d*(|em|[%*]))([lrc])/)
+            if md[1].to_i != 0
+              widthval = md[1]
+              widthval << "em" if md[2].empty?
+              widthopt = "width='#{widthval}' "
+            end
+            alignment = COLS_ALIGN[md[3]] || :left
           end
         end
-        res = inner(el, indent, opts)
         if alignment
-          "#{' '*indent}<ttcol #{widthopt}align='#{alignment}'#{el_html_attributes(el)}>#{res.empty? ? "&#160;" : res}</ttcol>\n"
-          else
+          res, irefs = clean_pcdata(inner_a(el, indent, opts))
+          warn "*** lost markup #{irefs} in table heading" unless irefs.empty?
+          "#{' '*indent}<ttcol #{widthopt}align='#{alignment}'#{el_html_attributes(el)}>#{res.empty? ? "&#160;" : res}</ttcol>\n" # XXX need clean_pcdata
+        else
+          res = inner(el, indent, opts)
           "#{' '*indent}<c#{el_html_attributes(el)}>#{res.empty? ? "&#160;" : res}</c>\n"
         end
       end
@@ -334,7 +384,10 @@ module Kramdown
         res = inner(el, indent, opts)
         target = el.attr['target']
         if target[0] == "#"     # handle [](#foo) as xref as in RFC 7328
-          el.attr['target'] = target[1..-1]
+          el.attr['target'] = target = target[1..-1]
+          if target.downcase == res.downcase
+            res = ''            # get rid of raw anchors leaking through
+          end
           "<xref#{el_html_attributes(el)}>#{res}</xref>"
         else
           "<eref#{el_html_attributes(el)}>#{res}</eref>"
@@ -356,15 +409,41 @@ module Kramdown
         fn = "#{REFCACHEDIR}/#{File.basename(url)}"
         f = File.stat(fn) rescue nil
         if !f || tn - f.ctime >= tvalid
-          $stderr.puts "#{fn}: #{f && tn-f.ctime}"
-          `cd #{REFCACHEDIR}; wget -t 3 -T 20 -Nnv "#{url}"` # ignore errors if offline (hack)
-          begin
-            File.utime nil, nil, fn
-          rescue Errno::ENOENT
-            warn "Can't fetch #{url} -- is wget in path?"
+          if f
+            message = "renewing (stale by #{"%.1f" % ((tn-f.ctime)/86400)} days)"
+            fetch_timeout = 10 # seconds, give up quickly if just renewing
+          else
+            message = "fetching"
+            fetch_timeout = 60 # seconds; long timeout needed for Travis
+          end
+          $stderr.puts "#{fn}: #{message}"
+          if ENV["HAVE_WGET"]
+            `cd #{REFCACHEDIR}; wget -t 3 -T #{fetch_timeout} -Nnv "#{url}"` # ignore errors if offline (hack)
+            begin
+              File.utime nil, nil, fn
+            rescue Errno::ENOENT
+              warn "Can't fetch #{url} -- is wget in path?"
+            end
+          else
+            require 'open-uri'
+            require 'timeout'
+            begin
+              Timeout::timeout(fetch_timeout) do
+                open(url) do |f|
+                  s = f.read
+                  if f.status[0] != "200"
+                    warn "*** Status code #{status} while fetching #{url}"
+                  else
+                    File.write(fn, s)
+                  end
+                end
+              end
+            rescue OpenURI::HTTPError, SocketError, Timeout::Error => e
+              warn "*** #{e} while fetching #{url}"
+            end
           end
         end
-        File.read(fn)
+        File.read(fn) # this blows up if no cache available after fetch attempt
       end
 
       XML_RESOURCE_ORG_MAP = {
@@ -421,6 +500,14 @@ module Kramdown
         content = inner(el.value, indent, opts).strip
         content = escape_html(content.sub(/\A<t>(.*)<\/t>\z/m) {$1}, :text) # text only...
         name = el.options[:name].sub(/\A[0-9]/) {"_" << $&}
+        while @footnote_names_in_use[name] do
+          if name =~ /:\d+\z/
+            name.succ!
+          else
+            name << ":1"
+          end
+        end
+        @footnote_names_in_use[name] = true
         attrstring = el_html_attributes_with(el, {"anchor" => name})
         "\n#{' '*indent}<cref#{attrstring}>#{content}</cref>"
       end
@@ -434,7 +521,8 @@ module Kramdown
 
       def convert_em(el, indent, opts)
         attrstring = el_html_attributes_with(el, {"style" => EMPH[el.type]})
-        "<spanx#{attrstring}>#{inner(el, indent, opts)}</spanx>"
+        span, irefs = clean_pcdata(inner_a(el, indent, opts))
+        "<spanx#{attrstring}>#{span}</spanx>#{irefs}"
       end
       alias :convert_strong :convert_em
 
@@ -468,26 +556,45 @@ module Kramdown
         "<#{type}#{el_html_attributes(el)}>#{escape_html(el.value, :text)}</#{type}>#{type == 'div' ? "\n" : ''}"
       end
 
+      ITEM_RE = '\s*(?:"([^"]*)"|([^,]*?))\s*'
+      IREF_RE = %r{\A(!\s*)?#{ITEM_RE}(?:,#{ITEM_RE})?\z}
+
+      def iref_attr(s)
+        md = s.match(IREF_RE)
+        attr = {
+          item: md[2] || md[3],
+          subitem: md[4] || md[5],
+          primary: md[1] && 'true',
+        }
+        "<iref#{html_attributes(attr)}/>"
+      end
+
+      def convert_iref(el, indent, opts)
+        iref_attr(el.attr['target'])
+      end
+
       def convert_abbreviation(el, indent, opts) # XXX: This is wrong
-        title = @doc.parse_infos[:abbrev_defs][el.value]
+        title = @root.options[:abbrev_defs][el.value]
         title = nil if title.empty?
-        "<abbr#{title ? " title=\"#{title}\"" : ''}>#{el.value}</abbr>"
+        value = el.value
+        if item = title
+          m = title.scan(Parser::RFC2629Kramdown::IREF_START)
+          if m.empty?
+            subitem = value
+          else
+            iref = m.map{|a,| iref_attr(a)}.join('')
+          end
+        else
+          item = value
+        end
+        iref ||= "<iref#{html_attributes(item: item, subitem: subitem)}/>"
+        "#{el.value}#{iref}"
       end
 
       def convert_root(el, indent, opts)
         result = inner(el, indent, opts)
       end
 
-      # Helper method for obfuscating the +text+ by using XML entities.
-      def obfuscate(text)
-        result = ""
-        text.each_byte do |b|
-          result += (b > 128 ? b.chr : "&#%03d;" % b)
-        end
-        result.force_encoding(text.encoding) if RUBY_VERSION >= '1.9'
-        result
-      end
-
     end
 
   end
